"""Bitwarden CSV import."""

import csv
from pathlib import Path
from typing import Dict, Any


class BitwardenCsv:
    """Write Bitwarden compatible CSV file."""

    def __init__(self, output_path: Path):
        self.file = output_path.open("w")
        self.writer = csv.writer(self.file, quoting=csv.QUOTE_MINIMAL)
        header = [
            "folder",
            "favorite",
            "type",
            "name",
            "notes",
            "fields",
            "reprompt",
            "login_uri",
            "login_username",
            "login_password",
            "login_totp",
        ]
        self.writer.writerow(header)

    def write_record(self, data: Dict[str, Any]) -> None:
        """Export data to CSV."""
        row = [
            data["folder"],
            "",  # favorite
            "login" if data["type"] == "card" else data["type"],
            data["name"],
            data["notes"],
            "\n".join([f"{field_name}: {value}" for field_name, value in data["fields"].items()]),
            "",  # reprompt
            data["login_uri"],
            data["login_username"],
            data["login_password"],
            "",  # login_totp
        ]
        self.writer.writerow(row)

    def close(self) -> None:
        """Close the CSV file."""
        self.file.close()



from bitwarden_import_msecure.__about__ import __version__
from bitwarden_import_msecure.main import bitwarden_import_msecure
from click.testing import CliRunner


def test_version():
    assert __version__


def test_version_option():
    runner = CliRunner()
    result = runner.invoke(bitwarden_import_msecure, ['--version'])
    assert result.exit_code == 0
    assert __version__ in result.output




from click.testing import CliRunner

from bitwarden_import_msecure.main import bitwarden_import_msecure


def assert_files_context_is_equal(file_path1, file_path2):
    """Compare the content of two files, abstracting away platform differences in newline characters."""
    with open(file_path1, 'r', newline=None, encoding='utf-8') as f1, open(file_path2, 'r', newline=None,
                                                                           encoding='utf-8') as f2:
        lines1 = f1.readlines()
        lines2 = f2.readlines()

        lines1 = [line.replace('\r\n', '\n').replace('\r', '\n') for line in lines1 if line.strip()]
        lines2 = [line.replace('\r\n', '\n').replace('\r', '\n') for line in lines2 if line.strip()]
        assert lines1 == lines2


def test_bitwarden_import_msecure_default_output(tmpdir, msecure_export, bitwarden_file):
    input_file = tmpdir.join("input.csv")
    input_file.write(msecure_export)

    runner = CliRunner()
    result = runner.invoke(bitwarden_import_msecure, [str(input_file)])
    assert result.exit_code == 0

    output_file = tmpdir.join("bitwarden.json")

    # bitwarden_file.write_text(output_file.read_text(encoding="utf8"))  # uncomment to refresh the expected output
    assert_files_context_is_equal(output_file, bitwarden_file)


def test_bitwarden_import_msecure_note_mode_default_output(tmpdir, msecure_export, bitwarden_notes_file):
    input_file = tmpdir.join("input.csv")
    input_file.write(msecure_export)

    runner = CliRunner()
    result = runner.invoke(bitwarden_import_msecure, [str(input_file), "--extra-fields", "notes"])
    assert result.exit_code == 0

    output_file = tmpdir.join("bitwarden.json")

    # bitwarden_notes_file.write_text(output_file.read_text(encoding="utf8"))  # uncomment to refresh the expected output
    assert_files_context_is_equal(output_file, bitwarden_notes_file)


def test_bitwarden_import_msecure_existing_output_file(tmpdir, msecure_export, bitwarden_file):
    input_file = tmpdir.join("input.txt")
    input_file.write(msecure_export)

    output_file = tmpdir.join("output.txt")
    output_file.write("existing data")

    runner = CliRunner()
    result = runner.invoke(bitwarden_import_msecure, [str(input_file), str(output_file)])
    assert result.exit_code == 1
    assert "Output file" in result.output and "already exists" in result.output
    assert result.exception
    assert isinstance(result.exception, SystemExit)
    assert result.exception.code == 1


def test_bitwarden_import_msecure_to_output_file(tmpdir, msecure_export, bitwarden_file):
    input_file = tmpdir.join("input.txt")
    input_file.write(msecure_export)

    output_file = tmpdir.join("output.txt")
    output_file.write("existing data")

    runner = CliRunner()
    result = runner.invoke(bitwarden_import_msecure, [str(input_file), str(output_file), "--force"])
    assert result.exit_code == 0

    # bitwarden_file.write_text(output_file.read_text(encoding="utf8"))  # uncomment to refresh the expected output
    assert_files_context_is_equal(output_file, bitwarden_file)
    assert input_file.read() == msecure_export  # Ensure input file remains unchanged


def test_bitwarden_import_msecure_default_csv_output(tmpdir, msecure_export, bitwarden_csv_file):
    input_file = tmpdir.join("input.csv")
    input_file.write(msecure_export)

    runner = CliRunner()
    result = runner.invoke(bitwarden_import_msecure, [str(input_file), "--format", "csv"])
    assert result.exit_code == 0

    output_file = tmpdir.join("bitwarden.csv")

    # bitwarden_csv_file.write_text(output_file.read_text(encoding="utf8"))  # uncomment to refresh the expected output
    assert_files_context_is_equal(output_file, bitwarden_csv_file)


def test_bitwarden_import_msecure_note_mode_default_csv_output(tmpdir, msecure_export, bitwarden_notes_csv_file):
    input_file = tmpdir.join("input.csv")
    input_file.write(msecure_export)

    runner = CliRunner()
    result = runner.invoke(bitwarden_import_msecure, [str(input_file), "--extra-fields", "notes", "--format", "csv"])
    assert result.exit_code == 0

    output_file = tmpdir.join("bitwarden.csv")

    # bitwarden_notes_csv_file.write_text(output_file.read_text(encoding="utf8"))  # uncomment to refresh the expected output
    assert_files_context_is_equal(output_file, bitwarden_notes_csv_file)



"""Write Bitwarden compatible JSON file."""

import json
import uuid
from pathlib import Path
from typing import Dict, Any
from datetime import datetime

BITWARDEN_TYPES = {"login": 1, "note": 2, "card": 3}


def now_string() -> str:
    """Return current time in Bitwarden format."""
    return datetime.now().astimezone().isoformat()


class BitwardenJson:
    """Write Bitwarden compatible JSON file."""

    data: Dict[str, Any]
    folder_ids: Dict[str, str]

    def __init__(self, output_path: Path):
        self.file = output_path
        self.data = {"encrypted": False, "folders": [], "items": []}
        self.folder_ids = {}

    @property
    def folders(self) -> Any:
        """Folder ID for the name.

        Creates a new folder if it doesn't exist.
        """
        parent = self

        class FolderManager:  # pylint: disable=too-few-public-methods
            """Dict like object."""

            def __getitem__(self, name: str) -> str:
                if name in parent.folder_ids:
                    return parent.folder_ids[name]
                folder_id = str(uuid.uuid4())
                parent.data["folders"].append({"id": folder_id, "name": name})
                parent.folder_ids[name] = folder_id
                return folder_id

        return FolderManager()

    def write_record(self, data: Dict[str, Any]) -> None:
        """Export data to JSON."""
        now = now_string()

        item = {
            "passwordHistory": None,
            "revisionDate": now,
            "creationDate": now,
            "deletedDate": None,
            "id": str(uuid.uuid4()),
            "organizationId": None,
            "folderId": self.folders[data["folder"]] if data["folder"] else None,
            "type": BITWARDEN_TYPES[data["type"]],
            "reprompt": 0,
            "name": data["name"],
            "notes": data["notes"],
            "favorite": False,
            "collectionIds": None,
        }
        if data["type"] == "login":
            item["login"] = {
                "fido2Credentials": [],
                "uris": [],
                "username": data["login_username"],
                "password": data["login_password"],
                "totp": None,
            }
        if data["type"] == "card":
            exp_month, exp_year = (
                data["fields"].pop("Expiration Date", "").split("/") + ["", ""]
            )[:2]
            cardholder_name = data["fields"].pop("Name on Card", "")
            if not cardholder_name:
                cardholder_name = data["fields"].pop("Name", "")
            item["card"] = {
                "cardholderName": cardholder_name,
                "brand": "",
                "number": data["login_username"],
                "expMonth": exp_month,
                "expYear": exp_year,
                "code": data["login_password"],
            }
        if data["type"] == "note":
            item["secureNote"] = {"type": 0}
        item["fields"] = [
            {
                "name": k,
                "value": v,
                "type": 1 if k in data["hidden_fields"] else 0,
                "linkedId": None,
            }
            for k, v in data["fields"].items()
        ]
        self.data["items"].append(item)

    def close(self) -> None:
        """Write the collected data to a JSON file."""
        with self.file.open("w", encoding="utf-8") as f:
            json.dump(self.data, f, indent=4)



"""mSecure export parser."""

from typing import Dict, List, Tuple, Any
import rich_click as click


BANK_FOLDER = "bank"


def import_msecure_row(row: List[str], extra_fields_to_notes: bool) -> Dict[str, Any]:
    """Extract data from mSecure row."""
    name = row[0].split("|")[0]
    if len(row[0].split("|")) > 2:
        print(f"Warning: name has more than one '|' character :`{row[0]}`.")
    record_type = "login"
    if row[1].strip() not in ["Login", "Credit Card", "Email Account"]:
        print(f"Warning: record type is not 'Login' :`{row[1]}`.")
    tag = row[2].strip()
    special_fields, fields, notes = extract_fields(row, extra_fields_to_notes)
    hidden_fields = []
    notes = "\n".join([part.replace("\\n", "\n") for part in [row[3], notes] if part.strip()])
    password, username = get_creds(special_fields, row)
    if special_fields["Card Number"]:
        if tag:
            click.echo(f"Warning: Tag `{tag}` present for Credit Card, ignored:\n{row}")
        else:
            tag = BANK_FOLDER
        record_type = "card"
    if "bank" in name.lower() and not tag:
        tag = BANK_FOLDER
    if not username and not password and not special_fields["Website"]:
        record_type = "note"
    if special_fields["PIN"]:
        fields["PIN"] = special_fields["PIN"]
        hidden_fields.append("PIN")

    return {
        "folder": tag,
        "type": record_type,
        "name": name,
        "notes": notes,
        "fields": fields,
        "hidden_fields": hidden_fields,  # names of fields to hide
        "login_uri": special_fields["Website"],
        "login_username": username,
        "login_password": password,
    }


def extract_fields(
    row: List[str], extra_fields_to_notes: bool
) -> Tuple[Dict[str, str], Dict[str, str], str]:
    """Extract fields from mSecure row.

    Return (special_fields, fields, notes)
    """
    special_fields = {
        "Website": "",
        "Username": "",
        "Password": "",
        "Card Number": "",
        "Security Code": "",
        "PIN": "",
    }
    fields = {}
    for field in row[4:]:
        parts = field.split("|")
        if parts[0] in special_fields:
            if special_fields[parts[0]]:
                print(f"Warning: Duplicate field `{parts[0]}` in row `{row}`.")
            special_fields[parts[0]] = "|".join(parts[2:])
        elif any(value.strip() for value in parts[2:]):
            fields[parts[0]] = ",".join(parts[2:])
    if extra_fields_to_notes:
        notes = "\n".join([f"{name}: {value}" for name, value in fields.items()])
        fields = {}
    else:
        notes = ""
    return special_fields, fields, notes


def get_creds(field_values: Dict[str, str], row: List[str]) -> Tuple[str, str]:
    """Get username and password."""
    username = field_values["Card Number"] or field_values["Username"]
    password = field_values["Security Code"] or field_values["Password"]
    if field_values["Card Number"] and field_values["Username"]:
        click.echo(f"Error: Both Card Number and Username present in row:\n{row}")
    if field_values["Security Code"] and field_values["Password"]:
        click.echo(f"Error: Both Security Code and Password present in row:\n{row}")
    return password, username
